# Written by Il-Chul Yoon <iyoon@cs.umd.edu>

package HPCALE_Client;

BEGIN{
    push (@INC, "$ENV{HPCALE_CLIENT_HOME}");
    push (@INC, "$ENV{HPCALE_CLIENT_HOME}/clientd");
    push (@INC, "$ENV{HPCALE_CLIENT_HOME}/client");
}

use util::Debug;
use validator;
use discovery;
use allocation;
use rtenv;
use execution;
use clientd;

use SOAP::Lite;
use IO::Socket;
use IPC::SysV qw(IPC_PRIVATE IPC_RMID S_IRWXU); 
use strict;

my $mod = "Launcher";
my $pipe= "$ENV{HPCALE_CLIENT_LOGPATH}/$ENV{HPCALE_CLIENT_LOGFILE}";

my $xjdfile = undef;
#my $host = undef;
#my $user = undef;
my $workdir = undef;
my $xjdprefix = undef;
my $jobid = undef;

Debug::logmsg($mod, $pipe, "\n");
Debug::logmsg($mod, $pipe, "=========================================");
Debug::logmsg($mod, $pipe, "     Client is started.");
Debug::logmsg($mod, $pipe, "=========================================");


# set soap info to access clientd 
my $clientd_uri = "http://$ENV{HPCALE_CLIENT_HOST}/FrontHandler";
my $clientd_proxy = "http://$ENV{HPCALE_CLIENT_HOST}:$ENV{HPCALE_CLIENT_PORT}";
my $soap = new SOAP::Lite(uri => $clientd_uri, proxy => $clientd_proxy);


# set shared memory information
my $buff = undef;
my $shmsize = 1000;
my $shm_id = shmget(IPC_PRIVATE, $shmsize, S_IRWXU) || die "$!";
if(!defined $shm_id){
   Debug::errmsg($mod, $pipe, "shared memory is not defined.");
   exit;
}

# retry means the number of times we can resleep to check whether
# the service is finished 
my $retry;
my $sleepunit; 
my $svc_result = 0;
my $keepgoing = 0;

# check command-line argument
&validation ;

# check env variable
&check_env_vars ; 

# fork clientd 
my $child_pid = undef;
if (!defined($child_pid = fork())){
   #error
   print "fork failed";
}
elsif ($child_pid){ # parent
   # 
   # wait clientd 
   # 
   &wait_clientd ; 

   # 
   # set basic shared information
   # 
   &set_sharedmem_id ; 
   &set_workdir ; 
   &set_xjdprefix ; 

   #
   # start asking services
   #
   &discovery or &abnormal_exit;
   &allocation or &abnormal_exit;
   &runtime_env or &abnormal_exit;
   &execution or &abnormal_exit;

   # 
   # terminate clientd
   # 
   &shutdown_clientd; 
}
else{ # child
   # 
   # start client daemon 
   # 
   &clientd::start_daemon();
}

sub validation 
{
   $xjdfile = validator::checkoption_x();
   #($host, $user, $workdir, $xjdprefix) = validator::validate($xjdfile);
   ($workdir, $xjdprefix) = validator::validate_x($xjdfile);
   $jobid = -1;
}

sub check_env_vars 
{
   my $chkresult = 1;

   Debug::logmsg($mod, "STDOUT", "Checking Environment Variables ...");
   my @envvars = qw/ HPCALE_CLIENT_HOST
                     HPCALE_CLIENT_HOME
                     HPCALE_CLIENT_PORT
                     HPCALE_CLIENT_LOGPATH
                     HPCALE_CLIENT_LOGFILE
                     HPCALE_DEBUG_LEVEL
                     HPCALE_CLIENT_SSLCERT
                     HPCALE_CLIENT_SSLKEY
                     HPCALE_SERVER_HOST
                     HPCALE_SERVER_PORT
                     HPCALE_SVC
                     HPCALE_FTPD_HOST
                     HPCALE_FTPD_PORT /;

   foreach(@envvars){
      if(chk_envvar($_) < 0){
         $chkresult = -1;
      }
   }

   if($chkresult < 0){
      Debug::logmsg($mod, "STDOUT", "Missing Environment Variable.");
      exit;
   }
   else{
      Debug::logmsg($mod, "STDOUT", "All required env. variables are defined.");
   }
}

sub chk_envvar {
   my $var = shift;
   if(!defined($ENV{$_})){
      Debug::logmsg($mod, "STDERR", "Undefined environment variable: $var");
      return -1;
   }
   else{
      return 1;
   }
}

sub shm_clear 
{
   shmwrite($shm_id, "\0", 0, 100) || return -1;
}

sub discovery 
{
   # XJD Refinement and Resource Discovery 
   my $r = discovery::discover($workdir, $xjdprefix);
   unless($r){
      Debug::errmsg($mod, $pipe, "Discovery failed.");
      Debug::errmsg($mod, "STDERR", "Discovery failed.");
      return 0;
   }

   # shared memory is written by clientd after receiving the discovery
   # result from server.  wait forever :)
   while(1){
      shmread($shm_id, $buff, 0, 100) || die "$!";
      substr($buff, index($buff, "\0")) = '';
      my $str = "discovery read : '$buff', length: " . length($buff)."\n";

      if(length($buff) == 0){
         Debug::dbgmsg(3, $mod, $pipe, "waiting XJD refinement...");
         `sleep 2`;
      }
      else{
         Debug::dbgmsg(3, $mod, $pipe, "$str");

         if(&check_shm($buff)){
            Debug::logmsg($mod, $pipe, "Discovery Service succeed.\n\n");
            Debug::logmsg($mod, "STDOUT", "Discovery Service succeed.\n\n");
         }
         else{
            Debug::errmsg($mod, $pipe, "Discovery Service failed.\n\n");
            Debug::errmsg($mod, "STDERR", "Discovery Service failed.\n\n");
            return 0;
         }

         &shm_clear();      
         last;
      }
   }

   return 1;
}

sub allocation 
{
   # Resource Allocation
   my $r = allocation::allocate($workdir, $xjdprefix);
   unless($r){
      Debug::errmsg($mod, $pipe, "Allocation failed.");
      return 0;
   }

   # shared memory is written by clientd after receiving the allocation
   # result from server.  wait forever :)
   while(1){
      shmread($shm_id, $buff, 0, 100) || die "$!";
      substr($buff, index($buff, "\0")) = '';
      my $str = "allocation read : '$buff', length: " . length($buff)."\n";

      if(length($buff) == 0){
         Debug::dbgmsg(3, $mod, $pipe, "waiting resource allocation...");
         `sleep 2`;
      }
      else{
         Debug::dbgmsg(3, $mod, $pipe, "$str");
         my @a = split /:/, $buff;
         $jobid = $a[2];

         Debug::dbgmsg(3, $mod, "STDOUT", "   Job ID: $a[2]");

         if(&check_shm($buff)){
            Debug::logmsg($mod, $pipe, "Allocation Service succeed.\n\n");
            Debug::logmsg($mod, "STDOUT", "Allocation Service succeed.\n\n");
         }
         else{
            Debug::errmsg($mod, $pipe, "Allocation Service failed.\n\n");
            Debug::errmsg($mod, "STDERR", "Allocation Service failed.\n\n");
            return 0;
         }

         &shm_clear();      
         last;
      }
   }

   return 1;
}

sub runtime_env
{
   # Runtime Environment Setup
   my $r = rtenv::setup($workdir, $xjdprefix, $jobid);
   unless($r){
      Debug::errmsg($mod, $pipe, "Runtime environment setup process failed.");
      return 0;
   }

   # shared memory is written by clientd after receiving the runtime environment
   # setup result from server.  wait forever :)
   while(1){
      shmread($shm_id, $buff, 0, 100) || die "$!";
      substr($buff, index($buff, "\0")) = '';
      my $str = "runtime_env read : '$buff', length: " . length($buff)."\n";

      if(length($buff) == 0){
         Debug::dbgmsg(3, $mod, $pipe, "waiting runtime env. setup");
         `sleep 2`;
      }
      else{
         Debug::dbgmsg(3, $mod, $pipe, "$str");

         if(&check_shm($buff)){
            Debug::logmsg($mod, $pipe, "Runtime Environment Preparation Service succeed.\n\n");
            Debug::logmsg($mod, "STDOUT", "Runtime Environment Preparation Service succeed.\n\n");
         }
         else{
            Debug::errmsg($mod, $pipe, "Runtime Environment Preparation Service failed.\n\n");
            Debug::errmsg($mod, "STDERR", "Runtime Environment Preparation Service failed.\n\n");
            return 0;
         }

         &shm_clear();      
         last;
      }
   }

   return 1;
}

sub execution
{
   # Component Execution
   my $r = execution::execute($workdir, $xjdprefix, $jobid);
   unless($r){
      Debug::errmsg($mod, $pipe, "Execution process failed.");
      return 0;
   }

   # shared memory is written by clientd after receiving the execution 
   # setup result from server.  wait forever :)
   while(1){
      shmread($shm_id, $buff, 0, 100) || die "$!";
      substr($buff, index($buff, "\0")) = '';
      my $str = "execution read : '$buff', length: " . length($buff)."\n";

      if(length($buff) == 0){
         Debug::dbgmsg(3, $mod, $pipe, "waiting component execution setup");
         `sleep 2`;
      }
      else{
         Debug::dbgmsg(3, $mod, $pipe, "$str");

         &check_shm($buff) or return 0;
         if(&check_shm($buff)){
            Debug::logmsg($mod, $pipe, "Execution Service succeed.\n\n");
            Debug::logmsg($mod, "STDOUT", "Execution Service succeed.\n\n");
         }
         else{
            Debug::errmsg($mod, $pipe, "Execution Service failed.\n\n");
            Debug::errmsg($mod, "STDERR", "Execution Service failed.\n\n");
            return 0;
         }

         &shm_clear();      
         last;
      }
   }

   return 1;
}

sub check_shm
{
   my $str = shift;
   my @a = split /:/, $str;
   ($a[1] eq "ok") ? return 1 : return 0;
}

sub set_sharedmem_id
{
   Debug::dbgmsg(1, $mod, $pipe, "Set shared memory id ($shm_id) at clientd");

   my $result = $soap->set_sharedmem_id($shm_id);
   unless($result->fault){
      Debug::logmsg($mod, $pipe, "sharedmem id set: ok");
   }
   else{
      my $err = join ', ', $result->faultcode, 
                           $result->faultstring, 
                           $result->faultdetail;
      Debug::logmsg($mod, $pipe, "Webservice fault : $err");
      Debug::logmsg($mod, $pipe, "SharedMem set failed in clientd.");
      &shutdown_clientd; 
      exit;
   }
}


sub set_workdir
{
   Debug::dbgmsg(1, $mod, $pipe, "Set client work dir at clientd");

   my $result = $soap->set_workdir($workdir);
   unless($result->fault){
      Debug::logmsg($mod, $pipe, "client workdir set: ok");
   }
   else{
      my $err = join ', ', $result->faultcode, 
                           $result->faultstring, 
                           $result->faultdetail;
      Debug::logmsg($mod, $pipe, "Webservice fault : $err");
      Debug::logmsg($mod, $pipe, "client workdir set failed in clientd.");
      &shutdown_clientd; 
      exit;
   }
}

sub set_xjdprefix
{
   Debug::dbgmsg(1, $mod, $pipe, "Set XJD prefix at clientd");

   my $result = $soap->set_xjdprefix($xjdprefix);
   unless($result->fault){
      Debug::logmsg($mod, $pipe, "XJD prefix set: ok");
   }
   else{
      my $err = join ', ', $result->faultcode, 
                           $result->faultstring, 
                           $result->faultdetail;
      Debug::logmsg($mod, $pipe, "Webservice fault : $err");
      Debug::logmsg($mod, $pipe, "XJD prefix set failed in clientd.");
      &shutdown_clientd; 
      exit;
   }
}

sub check_clientd
{
   my $flag = shift;
   Debug::logmsg($mod, $pipe, "checking client availability...");

   my $hostName = $ENV{HPCALE_CLIENT_HOST};
   my $host = shift || $hostName;
   my $port = shift || $ENV{HPCALE_CLIENT_PORT};

   while(1){
      my $sock = new IO::Socket::INET(PeerAddr=>$host,PeerPort=>$port,Proto=>'tcp');

      if($flag == 1){
         if($sock){
            close $sock or return -1;
            last;
         }
      }
      else{
         if(not defined $sock){
            last;
         }
      }

      `sleep 2`;
   }
   
   return 1;
}

sub wait_clientd
{
   my $r = &check_clientd(1);
   if($r < 0){
      Debug::logmsg($mod, $pipe, "Client Daemon is not available");
      &shutdown_clientd; 
      exit;
   }
   Debug::logmsg($mod, $pipe, "clientd is now available.");
}

sub shutdown_clientd
{
   Debug::logmsg($mod, $pipe, "shutting down clientd");
   my $result = $soap->shutdown_daemon();

   # check the real death of the clientd
   my $r = &check_clientd(-1);
   if($r < 0){
      Debug::logmsg($mod, $pipe, "unable to check clientd status. killing child.");
      kill 9 => $child_pid;
      exit;
   }
   Debug::logmsg($mod, $pipe, "ClientD is now turned off.");
}

sub abnormal_exit
{
   Debug::logmsg($mod, $pipe, "HPCALE exit due to failure.");
   Debug::logmsg($mod, "STDERR", "HPCALE exit due to failure.");
   &shutdown_clientd;
   exit;
}


