# Written by Il-Chul Yoon <iyoon@cs.umd.edu>

package HPCALE_Client;

BEGIN{
    push (@INC, "$ENV{HPCALE_CLIENT_HOME}");
    push (@INC, "$ENV{HPCALE_CLIENT_HOME}/clientd");
}

use util::Debug;
use validator;
use termination;
use clientd;
use SOAP::Lite;
use IO::Socket;
use IPC::SysV qw(IPC_PRIVATE IPC_RMID S_IRWXU); 
use strict;

my $mod = "JobStopper";

my $host = undef;
my $user = undef;
my $xjdfile = undef;
my $workdir = undef;
my $xjdprefix = undef;
my $jobid = undef;

Debug::logmsg($mod, "STDOUT", "\n");
Debug::logmsg($mod, "STDOUT", "=========================================");
Debug::logmsg($mod, "STDOUT", "     JobStopper is started.");
Debug::logmsg($mod, "STDOUT", "=========================================");


# set soap info to access clientd 
my $clientd_uri = "http://$ENV{HPCALE_CLIENT_HOST}/FrontHandler";
my $clientd_proxy = "http://$ENV{HPCALE_CLIENT_HOST}:$ENV{HPCALE_CLIENT_PORT}";
my $soap = new SOAP::Lite(uri => $clientd_uri, proxy => $clientd_proxy);


# set shared memory information
my $buff = undef;
my $shmsize = 1000;
my $shm_id = shmget(IPC_PRIVATE, $shmsize, S_IRWXU) || die "$!";
if(!defined $shm_id){
   Debug::errmsg($mod, "STDOUT", "shared memory is not defined.");
   exit;
}

# retry means the number of times we can resleep to check whether
# the service is finished 
my $retry;
my $sleepunit; 
my $svc_result = 0;
my $keepgoing = 0;

# check command-line argument
&validation ;

# check env variable
&check_env_vars ; 

# fork clientd 
my $child_pid = undef;
if (!defined($child_pid = fork())){
   #error
   print "fork failed";
}
elsif ($child_pid){ # parent
   # 
   # wait clientd 
   # 
   &wait_clientd ; 

   # 
   # set basic shared information
   # 
   &set_sharedmem_id ; 

   #
   # start asking service
   #
   &stopjob or &abnormal_exit;

   # 
   # terminate clientd
   # 
   &shutdown_clientd; 
}
else{ # child
   # 
   # start client daemon 
   # 
   &clientd::start_daemon();
}

sub validation 
{
   $jobid = validator::checkoption_j();
   $jobid = validator::validate_j($jobid);
}

sub check_env_vars 
{
   my $chkresult = 1;

   Debug::logmsg($mod, "STDOUT", "Checking Environment Variables ...");
   my @envvars = qw/ HPCALE_CLIENT_HOST
                     HPCALE_CLIENT_HOME
                     HPCALE_CLIENT_PORT
                     HPCALE_CLIENT_LOGPATH
                     HPCALE_CLIENT_LOGFILE
                     HPCALE_DEBUG_LEVEL
                     HPCALE_CLIENT_SSLCERT
                     HPCALE_CLIENT_SSLKEY
                     HPCALE_SERVER_HOST
                     HPCALE_SERVER_PORT
                     HPCALE_SVC
                     HPCALE_FTPD_HOST
                     HPCALE_FTPD_PORT /;

   foreach(@envvars){
      if(chk_envvar($_) < 0){
         $chkresult = -1;
      }
   }

   if($chkresult < 0){
      Debug::logmsg($mod, "STDOUT", "Missing Environment Variable.");
      exit;
   }
   else{
      Debug::logmsg($mod, "STDOUT", "All required env. variables are defined.");
   }
}

sub chk_envvar {
   my $var = shift;
   if(!defined($ENV{$_})){
      Debug::logmsg($mod, "STDERR", "Undefined environment variable: $var");
      return -1;
   }
   else{
      return 1;
   }
}

sub shm_clear 
{
   shmwrite($shm_id, "\0", 0, 100) || return -1;
}

sub stopjob 
{
   # XJD Refinement and Resource Discovery 
   my $r = termination::stopjob($jobid);
   unless($r){
      Debug::errmsg($mod, "STDOUT", "Stopping job failed.");
      return 0;
   }

   # shared memory is written by clientd after receiving the discovery
   # result from server.  wait forever :)
   while(1){
      shmread($shm_id, $buff, 0, 100) || die "$!";
      substr($buff, index($buff, "\0")) = '';
      my $str = "stopjob read : '$buff', length: " . length($buff)."\n";

      if(length($buff) == 0){
         Debug::dbgmsg(3, $mod, "STDOUT", "waiting to stop job...");
         `sleep 2`;
      }
      else{
         Debug::dbgmsg(3, $mod, "STDOUT", "$str");

         if(&check_shm($buff)){
            Debug::logmsg($mod, $pipe, "Job Termination Service succeed.\n\n");
            Debug::logmsg($mod, "STDOUT", "Job Termination Service succeed.\n\n");
         }
         else{
            Debug::errmsg($mod, $pipe, "Job Termination Service failed.\n\n");
            Debug::errmsg($mod, "STDERR", "Job Termination Service failed.\n\n");
            return 0;
         }

         &shm_clear();      
         last;
      }
   }

   return 1;
}

sub check_shm
{
   my $str = shift;
   my @a = split /:/, $str;
   ($a[1] eq "ok") ? return 1 : return 0;
}

sub set_sharedmem_id
{
   Debug::dbgmsg(1, $mod, "STDOUT", "Set shared memory id ($shm_id) at clientd");

   my $result = $soap->set_sharedmem_id($shm_id);
   unless($result->fault){
      Debug::logmsg($mod, "STDOUT", "sharedmem id set: ok");
   }
   else{
      my $err = join ', ', $result->faultcode, 
                           $result->faultstring, 
                           $result->faultdetail;
      Debug::logmsg($mod, "STDOUT", "Webservice fault : $err");
      Debug::logmsg($mod, "STDOUT", "SharedMem set failed in clientd.");
      &shutdown_clientd; 
      exit;
   }
}

sub check_clientd
{
   my $flag = shift;
   Debug::logmsg($mod, "STDOUT", "checking client daemon alive...");

   my $hostName = $ENV{HPCALE_CLIENT_HOST};
   my $host = shift || $hostName;
   my $port = shift || $ENV{HPCALE_CLIENT_PORT};

   while(1){
      my $sock = new IO::Socket::INET(PeerAddr=>$host,PeerPort=>$port,Proto=>'tcp');

      if($flag == 1){
         if($sock){
            close $sock or return -1;
            last;
         }
      }
      else{
         if(not defined $sock){
            last;
         }
      }

      `sleep 2`;
   }
   
   return 1;
}

sub wait_clientd
{
   my $r = &check_clientd(1);
   if($r < 0){
      Debug::logmsg($mod, "STDOUT", "Client Daemon is not available");
      &shutdown_clientd; 
      exit;
   }
   Debug::logmsg($mod, "STDOUT", "clientd is now available.");
}

sub shutdown_clientd
{
   Debug::logmsg($mod, "STDOUT", "shutting down clientd");
   my $result = $soap->shutdown_daemon();

   # check the real death of the clientd
   my $r = &check_clientd(-1);
   if($r < 0){
      Debug::logmsg($mod, "STDOUT", "unable to check clientd status. killing child.");
      kill 9 => $child_pid;
      exit;
   }
   Debug::logmsg($mod, "STDOUT", "ClientD is now turned off.");
}

sub abnormal_exit
{
   Debug::logmsg($mod, "STDOUT", "HPCALE exit due to failure.");
   &shutdown_clientd;
   exit;
}


