# Written by Il-Chul Yoon <iyoon@cs.umd.edu>

package FrontHandler;

use util::Debug;
use util::XJDWriter;
use Data::Dumper;
use FileHandler;
use strict;

my $mod = "ClientDaemon";
my $pipe= "$ENV{HPCALE_CLIENT_LOGPATH}/$ENV{HPCALE_CLIENT_LOGFILE}";
my $shm_id = undef;
my $workdir = undef;
my $xjdprefix = undef;

sub new 
{
   my $proto = shift;
   my $class = ref($proto) || $proto;
   my $self = {};

   bless $self, $class;
   return $self;
}

sub shutdown_daemon 
{
   my $self = shift;
   Debug::dbgmsg(3, $mod, $pipe, "Shutting down $mod (pid $$)");

   # send SIGUSR1 to my pid 
   kill INT => $$; 
}

sub set_sharedmem_id 
{
   my ($self, $id) = @_;
   $shm_id = $id;
   Debug::dbgmsg(3, $mod, $pipe, "Set shared memory: $shm_id");
}

sub set_workdir 
{
   my ($self, $wdir) = @_;
   $workdir = $wdir;
   Debug::dbgmsg(3, $mod, $pipe, "Client work directory: $workdir");
}

sub set_xjdprefix
{
   my ($self, $prefix) = @_;
   $xjdprefix = $prefix;
   Debug::dbgmsg(3, $mod, $pipe, "XJD prefix: $xjdprefix");
}

sub shm_write 
{
   my ($self, $stage, $msg) = @_;
   Debug::dbgmsg(3, $mod, $pipe, "At stage $stage, <$msg> is written to sharedmem");

   shmwrite($shm_id, "$stage:$msg", 0, 100) || return -1;
}

sub discoveryResult 
{
   my ($self, $lowXJD) = @_;
   my $stage = "discovery";
   my $msg = undef;

   # store the notified XJD file stream
   if(not defined $lowXJD){
      Debug::dbgmsg(3, $mod, $pipe, "received undefined low-level XJD.");
      return -1;
   }
   else{
      Debug::dbgmsg(3, $mod, $pipe, "\n".Dumper($lowXJD)."\n");
      Debug::dbgmsg(3, $mod, $pipe, "received undefined low-level XJD.");
      (new XJDWriter())->writeXML($lowXJD, "$workdir/$xjdprefix-low.xjd");
   }

   # sanity check
   if(not -e "$workdir/$xjdprefix-low.xjd"){
      Debug::dbgmsg(3, $mod, $pipe, "low-level XJD is not written normally.");
      $msg = "fail";
      $self->shm_write($stage, $msg);
   }
   else{
      Debug::dbgmsg(3, $mod, $pipe, "low-level XJD is written normally.");
      $msg = "ok";
      $self->shm_write($stage, $msg);
   }

   return 1;
}

sub allocationResult 
{
   my ($self, $jobid, $flag) = @_;
   Debug::dbgmsg(3, $mod, $pipe, "<JobID: $jobid><Flag: $flag>");

   my $stage = "allocation";
   my $msg = undef;

   if($flag < 0){ # reservation failure
      Debug::dbgmsg(3, $mod, $pipe, "received negative flag.");
      $msg = "fail";
      $self->shm_write($stage, $msg);
   }
   else{
      Debug::dbgmsg(3, $mod, $pipe, "received positive flag.");
      $msg = "ok:$jobid";
      $self->shm_write($stage, $msg);
   }

   return 1;
}

sub setupRTEnvResult 
{
   my ($self, $flag) = @_;

   Debug::dbgmsg(3, $mod, $pipe, "<Flag: $flag>");

   my $stage = "rtenvsetup";
   my $msg = undef;

   if($flag < 0){ # runtime environment setup failure
      Debug::dbgmsg(3, $mod, $pipe, "received negative flag.");
      $msg = "fail";
      $self->shm_write($stage, $msg);
   }
   else{
      Debug::dbgmsg(3, $mod, $pipe, "received positive flag.");
      $msg = "ok";
      $self->shm_write($stage, $msg);
   }

   return 1;
}

sub txfer_infiles 
{
   my ($self, $jobid, $specfiles) = @_;

   Debug::logmsg($mod, $pipe, "Got file transfer request from HPCALE server");
   Debug::dbgmsg(3, $mod, $pipe, "<JobID: $jobid>");
   Debug::dbgmsg(3, $mod, $pipe, "<SpecFiles: $specfiles>");

   $specfiles = FileHandler::pack_line($specfiles);

   if(length($specfiles) == 0){
      Debug::logmsg($mod, $pipe, "no need to transfer input files.");
      return 1;
   }
   else{
      Debug::logmsg($mod, $pipe, "start sending files.");

      if(-e "$workdir/$specfiles"){
         my $host = $ENV{HPCALE_FTPD_HOST};
         my $port = $ENV{HPCALE_FTPD_PORT};
         my $cert = $ENV{HPCALE_CLIENT_SSLCERT};

         FileHandler::contact_ftpd($host,$port,$jobid,$cert) or return 0;
       
         my @files = FileHandler::get_filelist($workdir, $specfiles);

         FileHandler::send_int(0); # this is to send input files 

         FileHandler::send_int($#files); # send num of files
         if($#files < 0){
            FileHandler::terminate_connection();
            return 1;
         }
            
         # send actual files
         my $r = undef; 
         foreach(@files){
            $r = FileHandler::send_file($_->[0], $_->[1], $_->[2], $_->[3]);
            if($r != 1){
               Debug::errmsg($mod, $pipe, "file transfer (or relocation) to resource failed.");
               last;
            }
            else{
               Debug::errmsg($mod, $pipe, "file $_->[1] is transferred to HPCALE Server.");
            }
         }
         
         FileHandler::terminate_connection();
         ($r == 1) ? return 1 : return 0;
      }
      else{
         Debug::errmsg($mod, $pipe, "no such file - $workdir/$specfiles ");
         return 0;
      }
   }

   return 0;
}

sub executionResult 
{
   my ($self, $flag) = @_;

   Debug::dbgmsg(3, $mod, $pipe, "<Flag: $flag>");

   my $stage = "execution";
   my $msg = undef;

   if($flag < 0){ # runtime environment setup failure
      Debug::dbgmsg(3, $mod, $pipe, "received negative flag.");
      $msg = "fail";
      $self->shm_write($stage, $msg);
   }
   else{
      Debug::dbgmsg(3, $mod, $pipe, "received positive flag.");
      $msg = "ok";
      $self->shm_write($stage, $msg);
   }

   return 1;
}

sub txfer_outfiles 
{
   my ($self, $jobid, $head, $specfiles, $stdout, $stderr) = @_;

   Debug::logmsg($mod, $pipe, "Got file receive request from HPCALE server");
   Debug::dbgmsg(3, $mod, $pipe, "<JobID: $jobid>");
   Debug::dbgmsg(3, $mod, $pipe, "<Head: $head>");
   Debug::dbgmsg(3, $mod, $pipe, "<SpecFiles: $specfiles>");
   Debug::dbgmsg(3, $mod, $pipe, "<STDOUT: $stdout>");
   Debug::dbgmsg(3, $mod, $pipe, "<STDERR: $stderr>");
   #Debug::dbgmsg(3, $mod, $pipe, "<HPCALELog: $hpcale_log>");

   $specfiles = FileHandler::pack_line($specfiles);

   Debug::logmsg($mod, $pipe, "start receiving files.");

   if(-e "$workdir/$specfiles"){
      my $host = $ENV{HPCALE_FTPD_HOST};
      my $port = $ENV{HPCALE_FTPD_PORT};
      my $cert = $ENV{HPCALE_CLIENT_SSLCERT};

      FileHandler::contact_ftpd($host,$port,$jobid,$cert) or return 0;
    
      my @files = FileHandler::get_filelist_outfile($workdir, $specfiles);

      # add stdout and stderr
      push @files, [$head, $stdout, "localhost", "$workdir/$stdout"];
      push @files, [$head, $stderr, "localhost", "$workdir/$stderr"];
      #push @files, [$head, $hpcale_log, "localhost", "$workdir/HPCALE_server.log"];

      FileHandler::send_int(1); # this is to receive output files 

      FileHandler::send_int($#files); # send num of files
      if($#files < 0){ # may never happen
         FileHandler::terminate_connection();
         return 1;
      }
         
      # recv actual files
      my $r = undef; 
      foreach(@files){
         $r = FileHandler::recv_file($_->[0], $_->[1], $_->[2], $_->[3], $workdir);
         if($r != 1){
            Debug::errmsg($mod, $pipe, "file transfer (or relocation) from resource failed.");
            last;
         }
      }

      FileHandler::terminate_connection();
      ($r == 1) ? return 1 : return 0;
   }
   else{
      Debug::errmsg($mod, $pipe, "no such file - $workdir/$specfiles ");
      return 0;
   }

   return 0;
}

#
# output is directed to STDOUT since stopjob service does not need 
# to write in the log file. (it returns just success/fail info.)
# 
sub stopjobResult
{
   my ($self, $flag) = @_;

   Debug::dbgmsg(3, $mod, "STDOUT", "<Flag: $flag>");

   my $stage = "stopjob";
   my $msg = undef;

   if($flag < 0){ # runtime environment setup failure
      Debug::dbgmsg(3, $mod, "STDOUT", "received negative flag.");
      $msg = "fail";
      $self->shm_write($stage, $msg);
   }
   else{
      Debug::dbgmsg(3, $mod, "STDOUT", "received positive flag.");
      $msg = "ok";
      $self->shm_write($stage, $msg);
   }

   return 1;
}

sub logmsg 
{
   my $self = shift;
   my $servmod = shift;
   my $msg = shift;

   Debug::logmsg($mod, $pipe, "logfrom server module($servmod): $msg");
    
   return 1;
}    
    
sub errmsg 
{
   my $self = shift;
   my $servmod = shift;
   my $msg = shift;

   # In this case, we may have something to do more 
   # to handle the error message. (although we write 
   # the message using logmsg subroutine
   Debug::logmsg($mod, $pipe, "errfrom server module($servmod): $msg");
    
   return 1;
}    
    
sub dbgmsg 
{
   my $self = shift;
   my $level = shift;
   my $servmod = shift;
   my $msg = shift;

   Debug::dbgmsg($level, $mod, $pipe, "dbgfrom server module($servmod): $msg");
    
   return 1;
}

1;

